/*
 * Copyright (c) 2016-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "acmrsim.h"

#include "mrsim_c_types.h"
#include "codels.h"


/* --- Function set_motor ----------------------------------------------- */

/** Codel mrsim_set_motor of function set_motor.
 *
 * Returns genom_ok.
 */
genom_event
mrsim_set_motor(double J, double b, double K, double R, double L,
                double V, const mrsim_s *mrsim,
                const genom_context self)
{
  struct mrsim_param_s *p = mrsim_params(mrsim);

  p->motor.J = J;
  p->motor.b = b;
  p->motor.K = K;
  p->motor.R = R;
  p->motor.L = L;
  p->motor.V = V;

  return genom_ok;
}


/* --- Function get_motor ----------------------------------------------- */

/** Codel mrsim_get_motor of function get_motor.
 *
 * Returns genom_ok.
 */
genom_event
mrsim_get_motor(double *J, double *b, double *K, double *R, double *L,
                double *V, const mrsim_s *mrsim,
                const genom_context self)
{
  struct mrsim_param_s *p = mrsim_params(mrsim);

  *J = p->motor.J;
  *b = p->motor.b;
  *K = p->motor.K;
  *R = p->motor.R;
  *L = p->motor.L;
  *V = p->motor.V;

  return genom_ok;
}


/* --- Function set_geom ------------------------------------------------ */

/** Codel mrsim_set_geom of function set_geom.
 *
 * Returns genom_ok.
 */
genom_event
mrsim_set_geom(const sequence_double *G, const double J[9],
               const mrsim_s *mrsim, const genom_context self)
{
  struct mrsim_param_s *p = mrsim_params(mrsim);
  uint16_t n;

  mrsim_get_rotors(&n, mrsim, self);
  if (G->_length != 6 * n) {
    errno = EDOM;
    return mrsim_e_sys_error("set_geom", self);
  }

  memcpy(p->body.G, G->_buffer, G->_length * sizeof(double));
  memcpy(p->body.J, J, 9 * sizeof(double));

  return genom_ok;
}


/* --- Function get_geom ------------------------------------------------ */

/** Codel mrsim_get_geom of function get_geom.
 *
 * Returns genom_ok.
 */
genom_event
mrsim_get_geom(sequence_double *G, double J[9], const mrsim_s *mrsim,
               const genom_context self)
{
  struct mrsim_param_s *p = mrsim_params(mrsim);
  uint16_t n, l;
  int i, j;

  mrsim_get_rotors(&n, mrsim, self);
  l = n * 6;
  genom_sequence_reserve(G, l);

  G->_length = l;
  for(i = 0; i < 6; i++)
    for(j = 0; j < n; j++)
      G->_buffer[i * n + j] = p->body.G[i * p->body.rotors + j];

  memcpy(J, p->body.J, 9 * sizeof(double));

  return genom_ok;
}


/* --- Function set_mass ------------------------------------------------ */

/** Codel mrsim_set_mass of function set_mass.
 *
 * Returns genom_ok.
 */
genom_event
mrsim_set_mass(double m, const mrsim_s *mrsim,
               const genom_context self)
{
  mrsim_params(mrsim)->body.m = m;
  return genom_ok;
}


/* --- Function get_mass ------------------------------------------------ */

/** Codel mrsim_get_mass of function get_mass.
 *
 * Returns genom_ok.
 */
genom_event
mrsim_get_mass(double *m, const mrsim_s *mrsim,
               const genom_context self)
{
  *m = mrsim_params(mrsim)->body.m;
  return genom_ok;
}


/* --- Function set_rotors ---------------------------------------------- */

/** Codel mrsim_set_rotors of function set_rotors.
 *
 * Returns genom_ok.
 * Throws mrsim_e_sys.
 */
genom_event
mrsim_set_rotors(uint16_t n, const mrsim_s *mrsim,
                 const genom_context self)
{
  struct mrsim_param_s *p = mrsim_params(mrsim);
  uint8_t i;

  if (n > p->body.rotors) {
    errno = EDOM;
    return mrsim_e_sys_error("set_rotors", self);
  }

  for(i = 0; i < p->body.rotors; i++)
    mrsim_motor(mrsim, i)->disabled = i < n ? false : true;

  return genom_ok;
}


/* --- Function get_rotors ---------------------------------------------- */

/** Codel mrsim_get_rotors of function get_rotors.
 *
 * Returns genom_ok.
 */
genom_event
mrsim_get_rotors(uint16_t *n, const mrsim_s *mrsim,
                 const genom_context self)
{
  struct mrsim_param_s *p = mrsim_params(mrsim);
  uint8_t i;

  for(i = p->body.rotors - 1; i >= 0; i--)
    if (!mrsim_motor(mrsim, i)->disabled) break;

  *n = i + 1;
  return genom_ok;
}
