/*
 * Copyright (c) 2016-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "acmrsim.h"

#include <err.h>
#include <math.h>
#include <time.h>

#include "mrsim_c_types.h"
#include "codels.h"


/* --- Task main -------------------------------------------------------- */


/** Codel mrsim_main_init of task main.
 *
 * Triggered by mrsim_start.
 * Yields to mrsim_main, mrsim_ether.
 */
genom_event
mrsim_main_init(mrsim_ids *ids, const mrsim_gps *gps,
                const mrsim_state *state, const genom_context self)
{
  or_pose_estimator_state *sdata, *gdata;
  struct mrsim_param_s *p;
  struct timespec tv;
  uint8_t i;

  ids->mrsim = mrsim_init(or_rotorcraft_max_rotors);
  if (!ids->mrsim) { warn("mrsim_init"); return mrsim_ether; }
  p = mrsim_params(ids->mrsim);

  clock_gettime(CLOCK_REALTIME, &tv);
  mrsim_update_time(ids->mrsim, tv);

  for(i = 0; i < p->body.rotors; i++)
    mrsim_motor(ids->mrsim, i)->disabled = i < 4 ? false : true;

  ids->gps_period = 0.1;

  sdata = state->data(self);
  double pvar = p->state.pstddev * p->state.pstddev;
  double qvar = p->state.qstddev * p->state.qstddev;
  double vvar = p->state.vstddev * p->state.vstddev;
  double wvar = p->state.wstddev * p->state.wstddev;
  double avar = p->state.astddev * p->state.astddev;

  *sdata = (or_pose_estimator_state){
    .ts = { 0, 0 },
    .intrinsic = false,

    .pos = {
      ._present = true,
      ._value = { .x = 0., .y = 0., .z = 0. }
    },
    .att = {
      ._present = true,
      ._value = { .qw = 1., .qx = 0., .qy = 0., .qz = 0. }
    },
    .pos_cov = {
      ._present = true,
      ._value = {{
          pvar,
          0., pvar,
          0.,   0., pvar,
        }}
    },
    .att_cov = {
      ._present = true,
      ._value = {{
          0.,
          0., qvar,
          0.,   0., qvar,
          0.,   0.,   0., qvar
        }}
    },
    .att_pos_cov = { ._present = false },

    .vel = {
      ._present = true,
      ._value = { .vx = 0., .vy = 0., .vz = 0. }
    },
    .vel_cov = {
      ._present = true,
      ._value = {{
          vvar,
          0., vvar,
          0.,   0., vvar,
        }}
    },

    .avel = {
      ._present = true,
      ._value = { .wx = 0., .wy = 0., .wz = 0. }
    },
    .avel_cov = {
      ._present = true,
      ._value = {{
          wvar,
          0., wvar,
          0.,   0., wvar
        }}
    },

    .acc = {
      ._present = true,
      ._value = {
        .ax = 0., .ay = 0., .az = 0.
      }
    },
    .acc_cov = {
      ._present = true,
      ._value = {{
          avar,
          0., avar,
          0.,   0., avar
        }}
    },

    .aacc = { ._present = false },
    .aacc_cov = { ._present = false }
  };

  gdata = gps->data(self);
  *gdata = (or_pose_estimator_state){
    .ts = sdata->ts,
    .intrinsic = false,

    .pos = sdata->pos,
    .att = { ._present = false },
    .pos_cov = sdata->pos_cov,
    .att_cov = { ._present = false },
    .att_pos_cov = { ._present = false },

    .vel = { ._present = false },
    .vel_cov = { ._present = false },

    .avel = { ._present = false },
    .avel_cov = { ._present = false },

    .acc = { ._present = false },
    .acc_cov = { ._present = false },

    .aacc = { ._present = false },
    .aacc_cov = { ._present = false }
  };

  return mrsim_main;
}


/** Codel mrsim_main_perm of task main.
 *
 * Triggered by mrsim_main.
 * Yields to mrsim_pause_main.
 */
genom_event
mrsim_main_perm(const mrsim_s *mrsim, bool *mrsim_lock,
                double gps_period, const mrsim_gps *gps,
                const mrsim_state *state, const genom_context self)
{
  or_pose_estimator_state *sdata = state->data(self);
  struct timespec tv;

  /* update time */
  clock_gettime(CLOCK_REALTIME, &tv);
  mrsim_update_time(mrsim, tv);

  /* update motors */
  mrsim_update_motors(mrsim);

  /* update state */
  mrsim_update_wrench(mrsim);
  mrsim_update_body(mrsim);
  mrsim_update_state(mrsim, sdata);

  /* update sensors */
  mrsim_update_imu(mrsim);

  gps->data(self)->pos = sdata->pos;
  gps->data(self)->pos_cov = sdata->pos_cov;
  mrsim_addnoise(mrsim, gps->data(self));


  /* publish or perish */
  mrsim_send_msgs(mrsim);

  sdata->ts.sec = tv.tv_sec;
  sdata->ts.nsec = tv.tv_nsec;
  state->write(self);

  if (sdata->ts.sec - gps->data(self)->ts.sec
      + 1e-9 * (sdata->ts.nsec - gps->data(self)->ts.nsec) >
      0.9 * gps_period) {
    gps->data(self)->ts = sdata->ts;
    gps->write(self);
  }

  return mrsim_pause_main;
}


/** Codel mrsim_main_fini of task main.
 *
 * Triggered by mrsim_stop.
 * Yields to mrsim_ether.
 */
genom_event
mrsim_main_fini(mrsim_s **mrsim, const genom_context self)
{
  mrsim_fini(*mrsim);
  return mrsim_ether;
}


/* --- Activity set_state ----------------------------------------------- */

/** Codel mrsim_set_state of activity set_state.
 *
 * Triggered by mrsim_start.
 * Yields to mrsim_ether.
 */
genom_event
mrsim_set_state(double x, double y, double z, double yaw, double pitch,
                double roll, double vx, double vy, double vz,
                double wx, double wy, double wz, const mrsim_s *mrsim,
                bool *mrsim_lock, const genom_context self)
{
  double cr = cos(roll/2.), cp = cos(pitch/2.), cy = cos(yaw/2.);
  double sr = sin(roll/2.), sp = sin(pitch/2.), sy = sin(yaw/2.);

  *mrsim_body(mrsim) = (struct mrsimstate_body_s) {
    .px = x, .py = y, .pz = z,
    .qw = cr * cp * cy + sr * sp * sy,
    .qx = sr * cp * cy - cr * sp * sy,
    .qy = cr * sp * cy + sr * cp * sy,
    .qz = cr * cp * sy - sr * sp * cy,

    .vx = vx, .vy = vy, .vz = vz,
    .wx = wx, .wy = wy, .wz = wz
  };

  return mrsim_ether;
}
